// Glenn Josefiak
// Fairfield University
// SW513
// Spring 2003

package j2d.gui;

// Standard libraries

import futils.StreamSniffer;
import gui.ClosableJFrame;
import gui.run.RunMenu;
import gui.run.RunMenuItem;
import j2d.*;
import j2d.file.ExtensionFileFilter;
import j2d.file.PPMToolkit;
import j2d.hpp.HistogramPanel;
import j2d.hpp.InvertFilter;
import j2d.imageproc.ExponentialStretchProcessor;
import j2d.imageproc.FalseColorProcessor;
import j2d.imageproc.HistogramEQProcessor;
import j2d.imageproc.LinearMappingProcessor;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.awt.image.PixelGrabber;
import java.io.*;
import java.util.Observable;
import java.util.Observer;

/**
 * This class provides a GUI for altering images
 */

public class Main extends ClosableJFrame
        implements ImageProcessListener {

    private MDIDesktopPane deskTop = new MDIDesktopPane();

    private FalseColorToolbox falseColorToolbox;
    private LinearMappingFrame linearMappingFrame;
    private HistogramEQToolbox histEqToolbox;
    private ExponentialStretchToolbox estStretchControls;


    // Image processors

    private FalseColorProcessor fcProcessor
            = new FalseColorProcessor();
    private LinearMappingProcessor lmProcessor
            = new LinearMappingProcessor();
    private HistogramEQProcessor heProcessor
            = new HistogramEQProcessor();
    private ExponentialStretchProcessor esProcessor
            = new ExponentialStretchProcessor();

    /**
     * Construct a new ImageTool
     */
    public Main() {
        Container c;

        c = getContentPane();
        c.setLayout(new BorderLayout());


        c.add(new JScrollPane(deskTop));

        setTitle("Image Tool");
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        // Set up the menus

        JMenuBar mb = new JMenuBar();
        setJMenuBar(mb);
        addFileMenu(mb);
        addProcessMenu(mb);
        WindowMenu windowMenu = new WindowMenu(deskTop);
        mb.add(windowMenu);
        windowMenu.setMnemonic('W');
    }

    private void addProcessMenu(JMenuBar mb) {
        RunMenu processMenu = new RunMenu("[process");
        mb.add(processMenu);
        processMenu.add(new RunMenuItem("[False Color") {
            public void run() {
                if (falseColorToolbox == null ||
                        falseColorToolbox.isClosed()) {
                    launchColorControls();
                }
            }
        });
        processMenu.add(new RunMenuItem("[negate{ctrl N}") {
            public void run() {
                InvertFilter invFilter = new InvertFilter();
                process(invFilter);

            }
        });
        processMenu.add(new RunMenuItem("salt and pepper 1000") {
            public void run() {
                Image img = getImage();
                ShortImageBean sib = new ShortImageBean(img);
                sib.saltAndPepper(1000);
                setImage(sib.getImage());
            }
        });
        processMenu.add(new RunMenuItem("gray") {
            public void run() {
                Image img = getImage();
                ShortImageBean sib = new ShortImageBean(img);
                sib.gray();
                setImage(sib.getImage());
            }
        });
        processMenu.add(new RunMenuItem("[Linear Mapping") {
            public void run() {
                if (linearMappingFrame == null ||
                        linearMappingFrame.isClosed()) {
                    launchMappingControls();
                }
            }
        });
        processMenu.add(new RunMenuItem("[Histogram EQ{ctrl H}") {
            public void run() {
                if (histEqToolbox == null ||
                        histEqToolbox.isClosed()) {
                    launchHistogramControls();
                }
            }
        });
        processMenu.add(new RunMenuItem("H[istogram Panel{ctrl i}") {
            public void run() {
                histogram();
            }
        });
        processMenu.add(new RunMenuItem("[Exponential Stretch") {
            public void run() {
                if (estStretchControls == null ||
                        estStretchControls.isClosed()) {
                    launchStretchControls();
                }
            }
        });
    }

    private void process(InvertFilter invFilter) {
        Image img = getImage();
        ShortImageBean sib = new ShortImageBean(img);

        sib.process(invFilter);
        setImage(sib.getImage());
    }

    private void histogram() {
        ClosableJFrame cf = new ClosableJFrame();

        Container c = cf.getContentPane();
        c.add(new HistogramPanel(this));
        cf.setSize(200, 200);
        cf.show();
    }

    public Image getImage() {
        ImageChildFrame icf = getChildFrame();
        return icf.getImage();
    }

    public Image getProcessedImage() {
        return getImage();
    }

    public void update(ImageProcessorInterface ip) {
        setImage(ip.process(getImage()));
    }

    public void setImage(Image img) {
        ImageChildFrame icf = getChildFrame();
        icf.setImage(img);
    }

    private ImageChildFrame getChildFrame() {
        ImageChildFrame icf = (ImageChildFrame)
                deskTop.getTopmostFrame(ImageChildFrame.class);
        return icf;
    }

    private void addFileMenu(JMenuBar mb) {
        RunMenu fileMenu = new RunMenu("[File");
        mb.add(fileMenu);

        fileMenu.add(new RunMenuItem("[Open{ctrl O}",
                Hex.getOpenIcon()) {
            public void run() {
                openImage();
            }
        });

        fileMenu.add(new RunMenuItem("[Capture screen{alt 4}",
                Hex.getCaptureIcon()) {
            public void run() {
                captureScreen();
            }
        });
        fileMenu.add(new RunMenuItem("C[apture camera{alt C}",
                Hex.getCaptureIcon()) {
            public void run() {
                captureVideo();
            }
        });
        fileMenu.add(new RunMenuItem("[Revert{ctrl W}",
                Hex.getRevertIcon()) {
            public void run() {
                revertImage();
            }
        });
        fileMenu.add(new RunMenuItem("[Save{ctrl S}",
                Hex.getSaveIcon()) {
            public void run() {
                saveImage();
            }
        });
        fileMenu.add(new RunMenuItem("[Print Hex{ctrl h}") {
            public void run() {
                printHexImage();
            }
        });

        fileMenu.add(new RunMenuItem("E[xit{ctrl Q}",
                Hex.getExitIcon()) {
            public void run() {
                shutdown();
            }
        });
    }

    public static String stripSuffix(String s) {
        int i = s.indexOf('.');
        if (i == -1) return s;
        return s.substring(0, i);
    }

    private void printHexImage() {

        String title = "foo";
        try {
            ImageChildFrame icf =
                    (ImageChildFrame) deskTop.getTopmostFrame(ImageChildFrame.class);
            if (icf == null) return;
            title = icf.getTitle();
        } catch (Exception e) {
            return;
        }
        Image img = getImage();
        if (img == null) System.out.println("img == null");
        ImageBean ib = new ImageBean();
        ib.setImage(img);
        int h = ib.getImageHeight();
        int w = ib.getImageWidth();

        String varName = stripSuffix(title);
        StringBuffer sb = new StringBuffer("   public static Icon get" +
                varName +
                "Icon() {\n" +
                "        return new ImageIcon(getImage(" +
                varName +
                ", " +
                h +
                ", " +
                w +
                "), \"" +
                varName +
                "\"" +
                ");\n" +
                "    }" +
                "\n\nprivate static int " +
                varName +
                "[] = {\n");
        Image img1 = ib.getImage();

        int pixels[] = new int[w * h];
        PixelGrabber pg = new PixelGrabber(img1,
                0,
                0,
                w,
                h,
                pixels,
                0,
                w);
        try {
            pg.grabPixels();
        } catch (InterruptedException e1) {
        }
        String newline = new String("\n");
        int i = 0;
        for (int y1 = 0; y1 < h; y1++) {
            for (int x1 = 0; x1 < w; x1++) {
                i = x1 + y1 * w;

                sb.append("0x" + Integer.toHexString(pixels[i]) + ", ");
            }
            sb.append(newline);

        }
        sb.append("};");
        String hexImage = sb.toString();
        System.out.println(hexImage);


    }

    /**
     * Helper function for closing the ImageTool.
     * Could put other termination code here.
     */
    private void shutdown() {
        dispose();
    }

    private void captureVideo() {

        CameraConfigToolbox cctb = new CameraConfigToolbox();
        cctb.show();


        VideoCaptureChildFrame vcf =
                new VideoCaptureChildFrame(cctb.getPushBufferDataSource(),
                        cctb.getSelectedFormat());
        vcf.show();
    }

    /**
     * This is called when the user requests a new file to be opened.
     */
    private void captureScreen() {

        try {
            BufferedImage bi = ImageUtils.captureWholeScreen();
            Image imgProcessee = ImageUtils.getImage(bi);

            ImageChildFrame icfNewFrame =
                    new ImageChildFrame(imgProcessee,
                            "screen shot");
            deskTop.add(icfNewFrame);
        } catch (AWTException e) {
            e.printStackTrace();

        }

    }

    /**
     * This is called when the user requests a new
     * file to be opened.
     */
    private void openImage() {
        File imageFile =
                futils.Futil.getReadFile("select an image file");
        Image imageToBeProcessed = getImage(imageFile);
        ImageChildFrame icfNewFrame =
                new ImageChildFrame(imageToBeProcessed,
                        imageFile.getName());
        deskTop.add(icfNewFrame);
    }

    public Image getImage(File f) {
        if (f == null) return null;
        if (!f.exists()) return null;

        InputStream is = null;
        try {
            is = new FileInputStream(f);
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }

        StreamSniffer ss = new StreamSniffer(is);
        int id = ss.classifyStream();
        switch (id) {
            case StreamSniffer.PPM:
                return getPPMImage(f);
            case StreamSniffer.PPM_RAWBITS:
                return getPPMImage(f);
            case StreamSniffer.GIF87a:
                return getGifOrJpg(f);
            case StreamSniffer.GIF89a:
                return getGifOrJpg(f);
            case StreamSniffer.JPEG:
                return getGifOrJpg(f);
            default:
                {
                    System.out.println("Can not open " +
                            ss +
                            " as image");
                    return null;
                }

        }
    }

    private Image getGifOrJpg(File imageFile) {
        return Toolkit.getDefaultToolkit()
                .getImage(imageFile.getAbsolutePath());
    }


    private Image getPPMImage(File imageFile) {
        try {
            Image imgProcessee;
            imgProcessee =
                    PPMToolkit.getImage(imageFile,
                            false);
            return imgProcessee;
        } catch (IOException e) {
            e.printStackTrace();
        }
        return null;
    }

    /**
     * This is called when the user picks the save
     * menu option.
     */
    private void saveImage() {
        ImageChildFrame icf;
        int intSuccess;
        File imageFile;


        try {
            icf =
                    (ImageChildFrame) deskTop.getTopmostFrame(ImageChildFrame.class);
            if (icf == null) return;
        } catch (Exception e) {
            return;
        }

        JFileChooser jfc = new JFileChooser();
        ExtensionFileFilter eff1 =
                new ExtensionFileFilter("ppm",
                        "Portable Pixelmap image file");
        jfc.addChoosableFileFilter(eff1);
        ExtensionFileFilter eff2 =
                new ExtensionFileFilter("ppm.gz",
                        "Zipped Pixelmap image file");
        jfc.addChoosableFileFilter(eff2);
        jfc.setFileFilter(eff1);
        jfc.setAcceptAllFileFilterUsed(false);

        intSuccess = jfc.showSaveDialog(this);

        if (intSuccess ==
                JFileChooser.APPROVE_OPTION) {
            imageFile = jfc.getSelectedFile();

            ExtensionFileFilter eff = (ExtensionFileFilter) jfc.getFileFilter();

            try {
                if (eff == eff1) {
                    imageFile =
                            fixExtension(imageFile,
                                    "ppm");
                    PPMToolkit.saveImage(icf.getImage(),
                            imageFile,
                            false);
                } else if (eff == eff2) {
                    imageFile =
                            fixExtension(imageFile,
                                    "ppm.gz");
                    PPMToolkit.saveImage(icf.getImage(),
                            imageFile,
                            true);
                }
                icf.setTitle(imageFile.getName());
            } catch (Exception e) {
                JOptionPane.showMessageDialog(this, "File write error: \n"
                        + e.toString());
            }
        }
    }

    /**
     * Fix the file extension
     */
    private File fixExtension(File file,
                              String extension) {
        String path = file.getAbsolutePath();

        if (!path.endsWith("." + extension)) {
            path += "." + extension;
        }
        return new File(path);
    }

    /**
     * Set the image in the active frame to the
     * base image.
     */
    private void revertImage() {
        ImageChildFrame icf;

        try {
            icf =
                    (ImageChildFrame)
                    deskTop.getTopmostFrame(ImageChildFrame.class);
            icf.revert();
        } catch (Exception e) {
            // do nothing
        }
    }

    /**
     * Helper function to show LinearMappingFrame
     */
    private void launchMappingControls() {
        linearMappingFrame =
                new LinearMappingFrame();
        linearMappingFrame.show();
        linearMappingFrame.getSliderBank()
                .addObserver(new Observer() {
                    public void update(Observable x,
                                       Object y) {
                        adjustLinearMapping(false);
                    }
                });
        linearMappingFrame.getButton()
                .addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        adjustLinearMapping(true);
                    }
                });
        deskTop.add(linearMappingFrame);
    }

    /**
     * Helper function to show FalseColorToolbox
     */
    private void launchColorControls() {
        falseColorToolbox =
                new FalseColorToolbox();
        falseColorToolbox.show();
        falseColorToolbox.getSliderBank()
                .addObserver(new Observer() {
                    public void update(Observable x,
                                       Object y) {
                        adjustColorization();
                    }
                });
        deskTop.add(falseColorToolbox);
    }

    /**
     * Helper function to show HistogramEQToolbox
     */
    private void launchHistogramControls() {
        histEqToolbox = new HistogramEQToolbox();
        histEqToolbox.show();
        histEqToolbox.getButton()
                .addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        adjustHistogram();
                    }
                });
        histEqToolbox.getSlider().addObserver(new Observer() {
            public void update(Observable x,
                               Object y) {
                adjustHistogram();
            }
        });

        deskTop.add(histEqToolbox);
    }

    /**
     * Helper function to show ExponentialStretchToolbox
     */
    private void launchStretchControls() {
        estStretchControls =
                new ExponentialStretchToolbox();
        estStretchControls.show();
        estStretchControls.getSlider()
                .addObserver(new Observer() {
                    public void update(Observable x,
                                       Object y) {
                        adjustExponentialStretch();
                    }
                });
        deskTop.add(estStretchControls);
    }


    /**
     * This is called each time a slider in the
     * FalseColorToolbox is moved. This is where
     * the real image processing happens.
     */
    private void adjustColorization() {
        float coeff[] = new float[3];
        ImageChildFrame icf;

        try {
            coeff =
                    falseColorToolbox.getSliderBank()
                    .getValues();

            icf =
                    (ImageChildFrame)
                    deskTop.getTopmostFrame(ImageChildFrame.class);
            fcProcessor.setBaseImage(icf.getBaseImage());
            fcProcessor.setColorization(coeff[0],
                    coeff[1],
                    coeff[2]);
            fcProcessor.performAlgorithm();
            icf.setImage(fcProcessor.getProcessedImage());
        } catch (Exception e) {
            // for now, do nothing.
        }
    }

    /**
     * This is called each time a slider on the
     * LinearMappingFrame is moved (btn = false)
     * or the button on the LinearMappingFrame
     * is pressed (btn = true).  This is where the
     * real image processing happens.
     */
    private void adjustLinearMapping(boolean btn) {
        float coeff[] = new float[2];
        ImageChildFrame icf;

        try {
            icf =
                    (ImageChildFrame) deskTop.getTopmostFrame(ImageChildFrame.class);
            lmProcessor.setBaseImage(icf.getBaseImage());
            if (btn) {
// compute parameters from the image, then update the toolbox sliders
                coeff =
                        lmProcessor.setOptimalParameters();
                linearMappingFrame.getSliderBank()
                        .setValues(coeff);
            } else {
// get the parameters from the sliders, then update the image
                coeff =
                        linearMappingFrame.getSliderBank()
                        .getValues();
                lmProcessor.setParameters(coeff[0], coeff[1]);
            }
            lmProcessor.performAlgorithm();
            icf.setImage(lmProcessor.getProcessedImage());
        } catch (Exception e) {
            // for now, do nothing.
        }
    }

    /**
     * This is called when the slider on the
     * HistogramEQToolbox is moved, or when its
     * button is pressed.
     */
    private void adjustHistogram() {
        boolean exponential;
        float alpha;
        ImageChildFrame icf;

        try {
            icf =
                    (ImageChildFrame) deskTop.getTopmostFrame(ImageChildFrame.class);
            heProcessor.setBaseImage(icf.getBaseImage());

            exponential =
                    histEqToolbox.getExponential();
            heProcessor.setExponential(exponential);
            if (exponential) {
                alpha =
                        histEqToolbox.getSlider()
                        .getValue();
                heProcessor.setAlpha(alpha);
            }

            heProcessor.performAlgorithm();
            icf.setImage(heProcessor.getProcessedImage());
        } catch (Exception e) {
            // for now, do nothing.
        }
    }

    /**
     * This is called each time a slider in the
     * ExponentialStretchToolbox is moved. This is
     * where the real image processing happens.
     */
    private void adjustExponentialStretch() {
        float coeff;
        ImageChildFrame icf;
        try {
            coeff =
                    estStretchControls.getSlider()
                    .getValue();

            icf =
                    (ImageChildFrame) deskTop.getTopmostFrame(ImageChildFrame.class);
            esProcessor.setBaseImage(icf.getBaseImage());
            esProcessor.setPower(coeff);
            esProcessor.performAlgorithm();
            icf.setImage(esProcessor.getProcessedImage());
        } catch (Exception e) {
            // for now, do nothing.
        }
    }

    /**
     * Entry point of the ImageTool application.
     */
    public static void main(String args[]) {
        Main fc = new Main();
        fc.setSize(500, 400);
        fc.show();
    }
}